;;; bhl.el --- From (B)rute text to (H)tml and (L)aTeX.

;; Copyright (C) 2002 2003 Bastien Guerry

;; Emacs Lisp Archive Entry
;; Filename: bhl.el
;; Author: Bastien Guerry <bastien1@free.fr>
;; Maintainer: Bastien Guerry <bastien1@free.fr>
;; Version: 1.3.3
;; Created: 16 november 2002
;; Revised: 01/24/2003
;; Keywords: convert plain text html latex sgml linuxdoc
;; Description: a major mode to convert raw text into HTML/LaTeX/SGML
;; URL: http://bhl.mode.free.fr/

;; This file is not part of GNU Emacs.

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2 of
;; the License, or (at your option) any later version.

;; This program is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY; without even the implied
;; warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
;; PURPOSE.  See the GNU General Public License for more details.

;; You should have received a copy of the GNU General Public
;; License along with this program; if not, write to the Free
;; Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
;; MA 02111-1307 USA

;;; Commentary:

;; BHL mode  is a mode  wich enables you  to convert plain  text files
;; into HTML, LaTeX and SGML  (Linuxdoc) files. This is a simple mode,
;; with simple conversion functions, dedicated to simple source files.

;;; Installation:

;; With `bhl.tar.gz':
;;   [your_shell]$ tar zxvf bhl-*.tar.gz
;;   [your_shell]$ make && make install
;;   [your_shell]$ make info && make install-info
;;   [your_shell]$ make pdf

;; With: `bhl.el':
;;   Just compile `bhl.el' and put it in your load-path.

;; Add to your `~/.emacs.el':
;;   (autoload 'bhl-mode "bhl" "BHL Mode" t)
;;   (add-to-list 'auto-mode-alist '("\\.bhl$" . bhl-mode))

;;; Usage:

;; bhl2html....................[C-c C-w]     -> convert BHL to HTML
;; bhl2latex...................[C-c C-l]     -> convert BHL to LaTeX
;; bhl2sgml....................[C-c C-s]     -> convert BHL to SGML (Linuxdoc)
;; bhl2txt.....................[C-c C-d]     -> convert BHL to TXT

;; bhl-view-log................[C-c C-o]     -> view log of the last conversion

;; bhl-insert-toc..............[C-c C-c t]   -> insert the table of contents
;; bhl-insert-url..............[C-c C-c h]   -> insert an URL
;; bhl-insert-img..............[C-c C-c i]   -> insert an image
;; bhl-insert-verbatim.........[C-c C-c v]   -> insert a verbatim environment
;; bhl-insert-mpage............[C-c C-c m]   -> insert a minipage

;; bhl-change-font-bold........[C-c C-f C-b] -> change font (bold)
;; bhl-change-font-emph........[C-c C-f C-e] -> change font (emphasis)
;; bhl-change-font-under.......[C-c C-f C-u] -> change font (underline)
;; bhl-change-font-tt..........[C-c C-f C-t] -> change font (truetype)
;; bhl-change-font-bolditalic..[C-c C-f C-_] -> change font (bold-italic)

;; bhl-show-toc................[C-c t]       -> generate a browsable table of contents
;; bhl-update-toc..............[C-c C-t]     -> update the numbering of sections

;; bhl-goto-next-section.......[C-c C-n]     -> go to the next section
;; bhl-goto-previous-section...[C-c C-p]     -> go to the previous section
;; bhl-goto-next-section.......[C-TAB]       -> go to the next URL
;; bhl-goto-previous-section...[C-S-TAB]     -> go to the previous URL

;; bhl-show-version............[C-c C-v]     -> display the BHL mode version number

;;; Known problems:

;; * No indentation of the output code.
;; * <LI> and <DD> are not followed by <P>.
;; * Don't end your source with a list item or a title of (sub)section
;; * BHL can't update section numbering if there is only subsection.
;; * BHL may convert special chars inside url/images.

;;; Acknowledgments

;; Thanks to Thierry Stoehr, Serge Basterot, Jason
;; Christoph Conrad, Peter Kindermann

;;; Summary:

;; Page 1  - Variables and constants
;; Page 2  - Custom
;; Page 3  - Faces
;; Page 4  - Mode
;; Page 5  - Menu
;; Page 6  - Miscellaneous
;; Page 7  - Table of content
;; Page 8  - Submit bug report
;; Page 9  - bhl2xxx

;;; Code:

(provide 'bhl)

(eval-when-compile
  (require 'footnote)
  (require 'skeleton)
  (require 'custom)
  (require 'cl)
  (require 'easymenu)
  (require 'derived)
  (require 'font-lock))


;;;; 1 - CONSTANTS

(defconst bhl-startup-message-lines
  '("Please use \\[bhl-submit-bug-report] to report bugs."
    "BHL comes with ABSOLUTELY NO WARRANTY."
    "Thanks for using the BHL mode!")
  "Lines to be displayed with `bhl-show-version'.")

(defconst bhl-help-address "bhl.mode@free.fr"
  "The address of the current maintainer.")

;;;; Regexp constants

(defconst bhl-generic-url-regexp
  "\\(https?://\\|file:///\\|telnet:\\|mime:\\|s?news:\\|ftp://\\|mailto:\\)\\([^ [\n\t]+\\>/?\\)"
  "Match a generic URL.")

(defconst bhl-url-regexp
  (concat "\\(\\(" bhl-generic-url-regexp "\\)\\|\\[\\[\\(\\(mailto:\\)?\\([^]\n\t]+\\>/?\\)\\)\\]\\[\\(\\<[^]\n\t]+\\)\\]\\]\\)")
  "Match a BHL URL.  Designed to work with `bhl-convert-url'.")

(defconst bhl-img-regexp
  "\\[\\[\\([^] \n\t]+\.\\(BMP\\|EPS\\|GIF\\|JP\\(?:E?G\\)\\|P\\(?:BM\\|GM\\|NG\\|PM\\|S\\)\\|TIFF\\|X\\(?:[BP]M\\)\\|bmp\\|eps\\|gif\\|jp\\(?:e?g\\)\\|p\\(?:bm\\|gm\\|ng\\|pm\\|s\\)\\|tiff\\|x\\(?:[bp]m\\)\\)\\)\\]\\]"
  "Match images.")	   ; `image-file-name-regexp' does not suffice

(defconst bhl-title-regexp
  "\\(\\`[ \t\n\r]*\\)\\([^\n]+\\)$"
  "Match the title.")

(defconst bhl-hr-regexp
  "^\\([<>-]\\)---+\\(\\[\\(\\([0-9]+\\)[^] ]*\\)\\]\\)?"
  "Match a horizontal rule.")

(defconst bhl-mpage-regexp
  "^\\([<>%]\\)\\([-%]\\)%+\\(\\[\\([^] ]+\\)\\]\\)?"
  "Match a starting minipage environment.")

(defconst bhl-list-regexp
  "^\t\\(\t\\)?\\(\t\\)?\\([-*o]\\|[0-9]+\.\\) "
  "Match lists items.")

(defconst bhl-descrip-regexp
  "\\(^\t\\)\\(\\([^\n-]+\\) -- \\)"
  "Match items of lists of descriptions.")

;;; Miscellaneous variables

(defvar bhl-version "1.3.3"
  "The current version of the BHL mode.")

(defvar bhl-main-buffer nil
  "The name of the main BHL buffer.")

(defvar bhl-menu nil
  "The BHL menu.")

(defvar bhl-point-list nil
  "The list of points used to browse the toc.")

(defvar bhl-tpl nil
  "The list of point positions that are required to browse the toc.")

(defvar bhl-tsl nil
  "The list of sections required to build and browse the toc.")

(defvar bhl-font-lock-keywords nil
  "Keywords to be fontified in the BHL mode.")

(defvar bhl-conversion-log nil
  "This variable stores the log for each conversion.")

(defvar bhl-latex-syntax-alist nil
  "Alist of syntactic elements for the `bhl2latex' conversion.")

;;;; Syntaxic variables and constants

(defvar bhl-xxx-syntax-alist nil
  "Alist of syntactic elements.")

(defvar bhl-xxx-conversions-list nil
  "Alist of conversion functions.")

(defconst bhl-syntax-list
  '("\\(__\\([^ \t\n\"_][^_]*\\)__\\)"
    "\\(==\\([^ \t\n=\"][^=]*\\)==\\)"
    "[^_]\\(_\\([^ \t\n_\"][^\t_]*\\)_\\)[^_]"
    "\\(\\*\\([^ \t\n*\"][^\t*]*\\)\\*\\)"
    "^[0-9]+\\. "
    "^[0-9]+\\.[0-9]+\\. "
    "^[0-9]+\\.[0-9]+\\.[0-9]+\\. ")
  "Syntactic elements understood by BHL.")

(defconst bhl-txt-syntax-alist
  '(("" "")
    ("" "")
    ("" "")
    ("" "")
    ("=== " " ===")
    ("== " " ==")
    ("= " " ="))
  "A list of TXT elements for the `bhl2txt' conversion.")

(defconst bhl-html-syntax-alist
  '(("<U>" "</U>")
    ("<TT>" "</TT>")
    ("<EM>" "</EM>")
    ("<B>" "</B>")
    ("<H1>" "</H1>")
    ("<H2>" "</H2>")
    ("<H3>" "</H3>"))
  "A list of HTML tags that correspond to the BHL syntax.")

(defconst bhl-sgml-syntax-alist
  '(("<IT>" "</IT>")		     ; no underline in sgml format (?)
    ("<TT>" "</TT>")
    ("<EM>" "</EM>")
    ("<BF>" "</BF>")
    ("<SECT>" "")
    ("<SECT1>" "")
    ("<SECT2>" ""))
  "A list of SGML tags that correspond to the BHL syntax.")

(defconst bhl-latex-syntax-num-alist
  '(("\\underline{" "}")
    ("\\texttt{" "}")
    ("\\emph{" "}")
    ("\\textbf{" "}")
    ("\\section{" "}")
    ("\\subsection{" "}")
    ("\\subsubsection{" "}"))
  "A list of LaTeX elements that correspond to the BHL syntax.")

(defconst bhl-latex-syntax-nonum-alist
  '(("\\underline{" "}")
    ("\\texttt{" "}")
    ("\\emph{" "}")
    ("\\textbf{" "}")
    ("\\section\*{" "}")
    ("\\subsection\*{" "}")
    ("\\subsubsection\*{" "}"))
  "Alist of LaTeX elements that correspond to the BHL syntax.")

;; Punctuation marks

(defconst bhl-fr-punctuation
  '(("\\(:\\)[ \t\n]" "&nbsp;:" "~:")
    ("\\(\\?\\)[ \t\n]" "&nbsp;?" "~?")
    ("\\(!\\)[ \t\n]" "&nbsp;!" "~!")
    ("\\(;\\)[ \t\n]" "&nbsp;;" "~;"))
  "A list of punctuation marks to convert according to the french typographic conventions.")

(defconst bhl-de-punctuation nil) ; no german punctuation conventions?

(defconst bhl-en-punctuation nil) ; no english punctuation conventions?

;; Quotation marks

(defconst bhl-en-quotation-marks
  '(("[ \n\t\(]\\(\"\\)" "&ldquo;" "``")
    ("[ \n\t\(]\\(\\)"  "&ldquo;" "``")
    ("\\(\\)[ \n\t,\.:;\?!\)]" "&rdquo;" "''")
    ("\\(\"\\)[ \n\t,\.:;\?!\)]" "&rdquo;" "''"))
  "A list of quote chars to convert.")

(defconst bhl-de-quotation-marks	; same as english?
  '(("[ \n\t\(]\\(\"\\)" "&ldquo;" "``")
    ("[ \n\t\(]\\(\\)"  "&ldquo;" "``")
    ("\\(\\)[ \n\t,\.:;\?!\)]" "&rdquo;" "''")	;; a vrifier
    ("\\(\"\\)[ \n\t,\.:;\?!\)]" "&rdquo;" "''"))
  "A list of quote chars to convert.")
    
(defconst bhl-fr-quotation-marks   ; warning: \usepackage[T1]{fontenc}
  '(("[ \n\t\(]\\(\"\\)" "&laquo;&nbsp;" "<<~")
    ("\\(\"\\)[ \n\t,\.:;\?!\)]" "&nbsp;&raquo;" "~>>")
    ("[ \n\t\(]\\(\\)" "&laquo;&nbsp;" "<<~")
    ("\\(\\)[ \n\t,\.:;\?!\)]" "&nbsp;&raquo;" "~>>"))
  "A list of quote characters to convert.")
    
;; Special characters

(defconst bhl-fr-special-chars
  '(("oe" "&oelig;" "\\oe{}")
    ("OE" "&OElig;" "\\OE{}")
    ("ae" "&aelig;" "\\ae{}")
    ("AE" "&AElig;" "\\AE{}"))
  "A list of strings to convert into ligatured characters.")

(defconst bhl-de-special-chars		; perhaps not so useful?
  '(("oe" "&ouml;" "")
    ("Oe" "&Ouml;" "")
    ("ae" "&auml;" "")
    ("Ae" "&Auml;" "")
    ("ue" "&uuml;" "")
    ("Ue" "&Uuml;" "")
    (""  "&szlig;" "\\ss{}"))
  "A list of strings to convert into ligatured characters.")

(defconst bhl-en-special-chars nil)   ; no english special characters?


;;;; 2 - CUSTOM

(defgroup bhl nil
  "BHL mode: convert plain TXT to HTML, LaTeX and SGML."
  :group 'wp
  :link '(info-link "(bhl)Top")
  :link '(url-link "http://bhl.mode.free.fr")
  :prefix "bhl-")

(defgroup bhl2html nil
  "Customize the `bhl2html' output."
  :group 'bhl)

(defgroup bhl2latex nil
  "Customize the `bhl2latex' output."
  :group 'bhl)

(defgroup bhl2sgml nil
  "Customize the `bhl2sgml' output."
  :group 'bhl)

(defgroup bhl2txt nil
  "Customize the `bhl2txt' output."
  :group 'bhl)

(defgroup bhl-toc nil
  "How to handle the table of contents."
  :group 'bhl)

(defgroup bhl-tables nil
  "How BHL must build tables."
  :group 'bhl)

(defgroup bhl-faces nil
  "*BHL faces used in BHL mode."
  :group 'faces
  :group 'bhl)

;;;; Main custom

(defcustom bhl-mode-hook 'footnote-mode
  "Normal hook run when entering BHL mode."
  :type 'hook
  :group 'bhl)

(defcustom bhl-author-name user-full-name
  "*The user's name."
  :type 'string
  :group 'bhl)

(defcustom bhl-i18n-conventions '("en" t t t)
  "English, french and german conventions.
These conventions are relative to the punctuation,
the quotation marks and some special characters."
  :type '(list (choice (const :tag "English" "en")
		       (const :tag "French" "fr")
		       (const :tag "German" "de"))
	       (boolean :tag "Punctuation     " :indent 2
			:help-echo "Non-nil means follow punctuation conventions.")
	       (boolean :tag "Quotation marks " :indent 2
			:help-echo "Non-nil means convert quotation marks.")
	       (boolean :tag "Special chars   " :indent 2
			:help-echo "Non-nil means convert special characters."))
  :group 'bhl)

(defcustom bhl-numbered-section-flag t
  "*Non-nil means that sections are numbered in the `bhl2latex' and `bhl2html' outputs."
  :type 'boolean
  :group 'bhl)

(defcustom bhl-check-tags-flag t
  "*Non-nil means BHL checks dubious tags before converting."
  :type 'boolean
  :group 'bhl)

(defcustom bhl-browse-url-function 'browse-url
  "Function to call to browse a URL."
  :type 'function
  :group 'bhl)

;;;; Customize the table of contents

(defcustom bhl-default-toc-depth 3
  "*The default depth of the table of contents."
  :type 'integer
  :group 'bhl-toc)

(defcustom bhl-intro-toc "--- Table of contents"
  "*A string inserted just before the table of contents."
  :type 'string
  :group 'bhl-toc)

(defcustom bhl-end-toc "--- End of the table of contents"
  "*A string inserted just after the table of contents."
  :type 'string
  :group 'bhl-toc)

(defcustom bhl-toc-location "top"
  "*The place into which the table of contents is inserted."
  :type '(choice (const :tag "Juste after title" "top")
		 (const :tag "Juste before end" "bottom")
		 (other :tag "On the point" "point"))
  :group 'bhl-toc)

;;;; Customize the tables

(defcustom bhl-table-location "htbp"
  "*The location of the table.
\"here top bottom page\" means put the table here or at the top
of the page or at the bottom of the page or on a separate page."
  :type '(choice (const :tag "here" "h")
		 (const :tag "top" "t")
		 (const :tag "bottom" "b")
		 (const :tag "page" "p")
		 (other :tag "here top bottom page" "htbp"))
  :group 'bhl-tables)

(defcustom bhl-table-align "center"
  "*How to align each table."
  :type '(choice (const "left")
		 (const "center")
		 (const "right"))
  :group 'bhl-tables)

(defcustom bhl-table-cell-align "c"
  "*How to align each cell in a table."
  :type '(choice (const :tag "left" "l")
		 (const :tag "center" "c")
		 (other :tag "right" "r"))
  :group 'bhl-tables)

(defcustom bhl-table-ask-caption-flag nil
  "*Non-nil means BHL asks for a caption in a table."
  :type 'boolean
  :group 'bhl-tables)

;;;; Customize the HTML output.

(defcustom bhl-html-conversions-list
  '(t t t t t t t t t t t)
  "*A list of conversion functions to perform with `bhl2html'.
Each element of this list correspond to a conversion function."
  :type '(list
	  (boolean :tag "Convert verbatim regions ")
	  (boolean :tag "Convert horizontal rules ")
	  (boolean :tag "Convert images           ")
	  (boolean :tag "Convert URLs             ")
	  (boolean :tag "Convert definition lists ")
	  (boolean :tag "Convert normal list      ")
	  (boolean :tag "Convert tables           ")
	  (boolean :tag "Convert LaTeX labels     "
		   :help-echo "Convert \"LaTeX\", \"LaTeX2e\" and \"TeX\" properly.")
	  (boolean :tag "Use i18n conventions     ")
	  (boolean :tag "Convert footnotes        ")
	  (boolean :tag "Convert minipage         "
		   :help-echo "Minipages are tables in HTML and minipages in LaTeX."))
  :group 'bhl2html)

(defcustom bhl-html-title-tags '("<H1 align=\"center\">" "</H1>")
  "A list of the opening and closing HTML tags for the title."
  :type '(list (string :tag "Opening tag: ")
	       (string :tag "Closing tag: "))
  :group 'bhl2html)

(defcustom bhl-html-meta-alist '(("generator" . "bhl2html"))
  "*A list of META tags.
The first field is the value of \"NAME\".
The second field is the value of \"CONTENT\".

Example:

author    (in the first field [key])
Monique   (in the second field [value])

outputs

<META NAME=\"author\" CONTENT=\"Monique\">"
  :type '(alist :key-type (string
			   :tag "Name"
			   :help-echo "Insert the \"NAME\" value of the META tag.")
		:value-type (string
			     :tag "Content"
			     :help-echo "Insert the \"CONTENT\" value of the META tag."))
  :group 'bhl2html)

(defcustom bhl-html-doctype
  "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">"
  "*The DOCTYPE description of the HTML document."
  :type 'string
  :group 'bhl2html)

(defcustom bhl-html-content-type
  "<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=iso-8859-1\">"
  "*The META tag which describes the content type of the HTML document."
  :type 'string
  :group 'bhl2html)

(defcustom bhl-html-style ""
  "*The style to be inserted in the header of the HTML output."
  :type 'string
  :group 'bhl2html)

(defcustom bhl-html-table-param-alist
  '(("cellpadding" . "3")
    ("cellspacing" . "0")
    ("border" . "1"))
  "*Alist of parameters for the conversion of tables.
The first field is the name of the parameter, the second field
specifies its value.

Example:

\"cellpadding\" (in the first field [key])
\"3\"   (in the second field [value])

outputs

<TABLE cellpadding=\"3\" ...>"
  :type '(alist :key-type (string
			   :tag "Property"
			   :help-echo "Insert the property name.")
		:value-type (string
			     :tag "Value"
			     :help-echo "Insert the value of the property."))
  :group 'bhl2html)

(defcustom bhl-html-para-align "none"
  "*How to align paragraphs."
  :type '(choice (const "left")
		 (const "center")
		 (const "justify")
		 (const "right")
		 (const "none"))
  :group 'bhl2html)

(defcustom bhl-html-img-align "center"
  "*How to align each image."
  :type '(choice (const "left")
		 (const "center")
		 (const "right"))
  :group 'bhl2html)

;;;; Customize the LaTeX output.

(defcustom bhl-latex-conversions-list
  '(t t t t t t t t t t t)
  "*A list of conversion functions to perform with `bhl2latex'.
Each element of this list corresponds to a conversion function."
  :type '(list
	  (boolean :tag "Convert verbatim regions ")
	  (boolean :tag "Convert horizontal rules ")
	  (boolean :tag "Convert images           ")
	  (boolean :tag "Convert URLs             ")
	  (boolean :tag "Convert definition lists ")
	  (boolean :tag "Convert normal list      ")
	  (boolean :tag "Convert tables           ")
	  (boolean :tag "Convert LaTeX labels     "
		   :help-echo "Convert \"LaTeX\", \"LaTeX2e\" and \"TeX\" properly.")
	  (boolean :tag "Use i18n conventions     ")
	  (boolean :tag "Convert footnotes        ")
	  (boolean :tag "Convert minipage         "
		   :help-echo "Minipages are tables in HTML and minipages in LaTeX."))
  :group 'bhl2latex)

(defcustom bhl-latex-packages-alist
  '(("inputenc" . "latin1"))
  "*Alist of packages to be included in the LaTeX header.
The first field is the name of the package, the second field
specifies optional elements.

Example:

inputenc (in the first field [key])
latin1   (in the second field [value])

outputs

\\usepackage[latin1]{inputenc}"
  :type '(alist :key-type (string
			   :tag "Package"
			   :help-echo "Insert the name of the LaTeX package.")
		:value-type (string
			     :tag "Options"
			     :help-echo "Insert the values of the package options."))
  :group 'bhl2latex)

(defcustom bhl-latex-class "article"
  "*The document's class for the LaTeX output."
  :type 'string
  :group 'bhl2latex)

(defcustom bhl-latex-pt 12
  "*The font height (in pt) for the LaTeX output."
  :type 'integer
  :group 'bhl2latex)

;; Customize the SGML output

(defcustom bhl-sgml-conversions-list
  '(t t t t t t t t nil t t)
  "*A list of conversion functions to perform with `bhl2sgml'.
Each element of this list corresponds to a conversion function."
  :type '(list
	  (boolean :tag "Convert verbatim regions ")
	  (boolean :tag "Convert horizontal rules ")
	  (boolean :tag "Convert images           ")
	  (boolean :tag "Convert URLs             ")
	  (boolean :tag "Convert definition lists ")
	  (boolean :tag "Convert normal list      ")
	  (boolean :tag "Convert tables           ")
	  (boolean :tag "Convert LaTeX labels     "
		   :help-echo "Convert \"LaTeX\", \"LaTeX2e\" and \"TeX\" properly.")
	  (boolean :tag "Use i18n conventions     ")
	  (boolean :tag "Convert footnotes        ")
	  (boolean :tag "Convert minipage         "
		   :help-echo "Minipages are tables in HTML and minipages in LaTeX."))
  :group 'bhl2sgml)

;; Customize the TXT output

(defcustom bhl-txt-conversions-list
  '(t t t t nil nil nil nil nil nil t)
  "*A list of conversion functions to perform with `bhl2html'.
Each element of this list correspond to a conversion function."
  :type '(list
	  (boolean :tag "Convert verbatim regions ")
	  (boolean :tag "Convert horizontal rules ")
	  (boolean :tag "Convert images           ")
	  (boolean :tag "Convert URLs             ")
	  (boolean :tag "Convert definition lists ")
	  (boolean :tag "Convert normal list      ")
	  (boolean :tag "Convert tables           ")
	  (boolean :tag "Convert LaTeX labels     "
		   :help-echo "Convert \"LaTeX\", \"LaTeX2e\" and \"TeX\" properly.")
	  (boolean :tag "Use i18n conventions     ")
	  (boolean :tag "Convert footnotes        ")
	  (boolean :tag "Convert minipage         "
		   :help-echo "Minipages are tables in HTML and minipages in LaTeX."))
  :group 'bhl2txt)

(defcustom bhl-txt-sec-center-flag nil
  "Non-nil means that BHL centers the sections in the TXT output."
  :type 'boolean
  :group 'bhl2txt)


;;;; 3 - FACES

(defface bhl-mode-line-face
  '((t (:foreground "blue")))
  "*The BHL face for some mode-line informations."
  :group 'bhl-faces)

(defface bhl-img-face
  '((t (:foreground "green4")))
  "*The BHL face for images."
  :group 'bhl-faces)

(defface bhl-url-face
  '((t (:foreground "green4")))
  "*The BHL face for URL."
  :group 'bhl-faces)

(defface bhl-list-face
  '((t (:foreground "blue")))
  "*The BHL face for list separators."
  :group 'bhl-faces)

(defface bhl-descrip-face
  '((t (:foreground "blue")))
  "*The BHL face for descriptive list separators."
  :group 'bhl-faces)

(defface bhl-hr-face
  '((t (:foreground "red")))
  "*The BHL face for horizontal rule."
  :group 'bhl-faces)

(defface bhl-toc-face
  '((t (:foreground "blue")))
  "*The BHL face for the table of contents separators."
  :group 'bhl-faces)

(defface bhl-title-face
  '((t (:weight bold :height 1.9)))
  "*The BHL face for the title."
  :group 'bhl-faces)

(defface bhl-section-face
  '((t (:weight bold :height 1.6)))
  "*The BHL face for the sections."
  :group 'bhl-faces)

(defface bhl-subsection-face
  '((t (:weight bold :height 1.3)))
  "*The BHL face for the subsections."
  :group 'bhl-faces)

(defface bhl-subsubsection-face
  '((t (:weight bold)))
  "*The BHL face pour les subsubsections."
  :group 'bhl-faces)

(defface bhl-tt-face
  '((t (:foreground "red")))
  "*The BHL face for truetype font text."
  :group 'bhl-faces)

(defface bhl-special-face
  '((t (:foreground "orange")))
  "*The BHL face for minipage and verbatim environment."
  :group 'bhl-faces)


;;;; 4 - MODE

(unless (fboundp 'match-string-no-properties)
  (defalias 'match-string-no-properties 'match-string))

(setq bhl-font-lock-keywords
      `((,bhl-title-regexp . 'bhl-title-face)
	(,bhl-list-regexp . (0 'bhl-list-face append))
	(,bhl-img-regexp . 'bhl-img-face)
	(,bhl-hr-regexp . 'bhl-hr-face)
	(,bhl-mpage-regexp . 'bhl-special-face)
	("^--$" . 'bhl-special-face)
	(,bhl-intro-toc . 'bhl-toc-face)
	(,bhl-end-toc . 'bhl-toc-face)
	(,(concat (nth 6 bhl-syntax-list) ".*$") . 'bhl-subsubsection-face)
	(,(concat (nth 5 bhl-syntax-list) ".*$") . 'bhl-subsection-face)
	(,(concat (nth 4 bhl-syntax-list) ".*$") . 'bhl-section-face)
	(,(nth 0 bhl-syntax-list) . (1 'underline append))
	(,(nth 1 bhl-syntax-list) . (1 'bhl-tt-face append))
	(,(nth 2 bhl-syntax-list) . (1 'italic append))
	(,(nth 3 bhl-syntax-list) . (1 'bold append))
	(,bhl-descrip-regexp . (0 'bhl-descrip-face append))
	bhl-url-highlight
	bhl-generic-url-highlight))

;;;###autoload
(define-derived-mode bhl-mode text-mode "BHL"
  "B - Brute, H - HTML, L - LaTeX.

BHL mode is a mode which enables you to convert plain text files into
HTML, LaTeX and SGML files.  This is a simple mode, with simple
conversion functions and converts only simple source files.
\\<bhl-mode-map>
+ Sections:
===========

1.<SPC>     : insert a section.
1.1.<SPC>   : insert a subsection.
1.1.1.<SPC> : insert a subsubsection.

+ Font change:
==============

*a word*   : bold
_a word_   : emphasis
_*a word*_ : bold-italic
__a word__ : underline
==a word== : truetype

--
This part of text will be inserted as verbatim text.
--

%%[2cm]
This part of text will be put inside a minipage
which width is 2cm (size is optional).
%%

+ Key bindings:
===============

\\[bhl2html] : convert the buffer into HTML.
\\[bhl2latex] : convert the buffer into LaTeX.
\\[bhl2sgml] : convert the buffer into SGML (Linuxdoc).
\\[bhl2txt] : convert the buffer into TXT.

\\[bhl-insert-url] : ask for an URL and its name, then insert it.
\\[bhl-insert-img] : ask for an image file name, then insert it.
\\[bhl-insert-hr] : insert a horizontal rule.
\\[bhl-insert-mpage] : insert a minipage.
\\[bhl-insert-verbatim] : insert a verbatim environment.
\\[bhl-insert-toc] : insert the table of contents.
   Take the toc depth as an optional argument.

\\[bhl-view-log] : view the log of the last conversion.
\\[bhl-compose-mail] : compose mail with the result of bhl2txt on the current buffer.

\\[bhl-show-toc] : browse the table of contents.
\\[bhl-update-toc] : update sections numbering.

\\[bhl-goto-next-section] : go to the next section.
\\[universal-argument] \\[bhl-goto-next-section] : go to the next section or (sub)subsection.
\\[bhl-goto-previous-section] : go to the previous section.
\\[universal-argument] \\[bhl-goto-previous-section] : go to the previous section or (sub)subsection.

\\[bhl-goto-next-url] : go to the next URL.
\\[bhl-goto-previous-url] : go to the previous URL.

\\[bhl-submit-bug-report] : submit a bug report.
\\[bhl-show-version] : display the current version of BHL.

+ List separators:
==================

TAB(S) and \"*\" or \"-\" or \"o\" and a whitespace: a non-ordered list item.
TAB(S) and 0-9 and dot and a whitespace: an ordered list item.
TAB and a definition term and \"--\" and a whitespace: a definition list item.

Only ordered and non-ordered lists can be nested.
The limit of nesting depth is set to 3.

+ List examples:
================

   * Here is
      - a simple
      - nested list
         o with three levels
         o of nested items.
      1. Here is...
      2. ...an ordered list
   * inside another list.

   A term -- and its definition.
   Another term -- and its definition.

+ bhl-mode-map:
===============

\\{bhl-mode-map}"
  (kill-all-local-variables)
  (auto-fill-mode 1)
  (use-local-map bhl-mode-map)
  (make-local-variable 'font-lock-defaults)
  (easy-menu-add bhl-menu)
  (setq bhl-main-buffer (buffer-name)
		  imenu-generic-expression '((nil "^\\([0-9]+\.\\)+ \\(.*$\\)" 2))
		  skeleton-end-newline nil
		  tab-width 3
		  font-lock-defaults '(bhl-font-lock-keywords t)
		  font-lock-multiline t
		  mode-name "BHL"
		  major-mode 'bhl-mode)
  (setq mode-line-format
	(list mode-line-modified
	      mode-line-buffer-identification " "
	      global-mode-string
	      "   %[("
	      'mode-name
	      'mode-line-process
	      'minor-mode-alist
	      "%n" ")%]"
	      (propertize " [BHL: " 'face '(t bhl-mode-line-face))
	      (if bhl-numbered-section-flag "Sec num" "No sec num")
	      ", "
	      (if bhl-table-ask-caption-flag "caption" "no caption")
	      ", toc at "
	      'bhl-toc-location
	      ", " (car bhl-i18n-conventions)
	      "]--"
	      '(line-number-mode "L%l--")
	      "%p" "-%-"))
  (run-hooks 'bhl-mode-hook))

;; Conversion commands
(define-key bhl-mode-map "\C-c\C-w" 'bhl2html)
(define-key bhl-mode-map "\C-c\C-l" 'bhl2latex)
(define-key bhl-mode-map "\C-c\C-s" 'bhl2sgml)
(define-key bhl-mode-map "\C-c\C-d" 'bhl2txt)

;; Insert
(define-key bhl-mode-map "\C-c\C-ch" 'bhl-insert-url)
(define-key bhl-mode-map "\C-c\C-ci" 'bhl-insert-img)
(define-key bhl-mode-map "\C-c\C-cm" 'bhl-insert-mpage)
(define-key bhl-mode-map "\C-c\C-cv" 'bhl-insert-verbatim)
(define-key bhl-mode-map "\C-c\C-cr" 'bhl-insert-hr)

;; Font commands
(define-key bhl-mode-map "\C-c\C-f\C-b" 'bhl-change-font-bold)
(define-key bhl-mode-map "\C-c\C-f\C-e" 'bhl-change-font-emph)
(define-key bhl-mode-map "\C-c\C-f\C-u" 'bhl-change-font-under)
(define-key bhl-mode-map "\C-c\C-f\C-t" 'bhl-change-font-tt)
(define-key bhl-mode-map "\C-c\C-f\C-_" 'bhl-change-font-bolditalic)

;; Toc commands
(define-key bhl-mode-map "\C-c\C-ct" 'bhl-insert-toc)
(define-key bhl-mode-map "\C-c\C-t" 'bhl-update-toc)
(define-key bhl-mode-map "\C-ct" 'bhl-show-toc)
(define-key bhl-mode-map [S-iso-lefttab] 'bhl-tab)

;; Move inside your document
(define-key bhl-mode-map "\C-c\C-n" 'bhl-goto-next-section)
(define-key bhl-mode-map "\C-c\C-p" 'bhl-goto-previous-section)
(define-key bhl-mode-map [C-tab] 'bhl-goto-next-url)
(define-key bhl-mode-map [C-S-iso-lefttab] 'bhl-goto-previous-url)

;; Other
(define-key bhl-mode-map "\C-c\C-v" 'bhl-show-version)
(define-key bhl-mode-map "\C-c\C-o" 'bhl-view-log)
(define-key bhl-mode-map "\C-c\C-xm" 'bhl-compose-mail)
(fset 'bhl 'bhl-mode)


;;;; 5 - MENU

(easy-menu-define bhl-menu bhl-mode-map
  "Menu of the BHL mode"
  '("BHL"
    ;; the main conversion commands
    ("Convert"
     ["bhl2html" bhl2html :help "Convert into HTML."]
     ["bhl2latex" bhl2latex :help "Convert into LaTeX."]
     ["bhl2sgml" bhl2sgml :help "Convert into SGML (Linuxdoc)."]
     ["bhl2txt" bhl2txt :help "Convert into TXT."])
    ;; insert commands
    ("Insert"
     ["Insert TOC" bhl-insert-toc t]
     ["Insert URL" bhl-insert-url t]
     ["Insert image" bhl-insert-img t]
     ["Insert minipage" bhl-insert-mpage t]
     ["Insert verbatim" bhl-insert-verbatim t]
     ["Insert hr" bhl-insert-hr t])
    ;; changing font commands
    ("Change font"
     ["Bold" bhl-change-font-bold t]
     ["Emphasis" bhl-change-font-emph t]
     ["Underline" bhl-change-font-under t]
     ["TrueType" bhl-change-font-tt t]
     ["Bold italic" bhl-change-font-bolditalic t])
    ;; browse the source
    ("Go to"
     ["Next section" bhl-goto-next-section t]
     ["Previous section" bhl-goto-previous-section t]
     ["Next URL" bhl-goto-next-url t]
     ["Previous URL" bhl-goto-previous-url t])
    "---"
    ;; handling the table of contents
    ["Update TOC" bhl-update-toc :help "Update sections numbering."]
    ["Browse the TOC" bhl-show-toc :help "Browse the TOC in a new buffer."]
    ("TOC location"
     ["TOC at the top" (progn (setq bhl-toc-location "top") (bhl-mode))
      :style radio :selected (equal bhl-toc-location "top")]
     ["TOC at the bottom" (progn (setq bhl-toc-location "bottom") (bhl-mode))
      :style radio :selected (equal bhl-toc-location "bottom")]
     ["TOC on the point" (progn (setq bhl-toc-location "point") (bhl-mode))
      :style radio :selected (equal bhl-toc-location "point")])
    ("TOC depth"
     ["Sections only" (setq bhl-default-toc-depth 1)
      :style radio :selected (equal bhl-default-toc-depth 1)]
     ["Sections and subsections" (setq bhl-default-toc-depth 2)
      :style radio :selected (equal bhl-default-toc-depth 2)]
     ["Maximum depth [3]" (setq bhl-default-toc-depth 3)
      :style radio :selected (equal bhl-default-toc-depth 3)])
    "---"
    ;; miscellaneous
    ("I18n conventions"
     ["English" (progn (setcar bhl-i18n-conventions "en") (bhl-mode))
      :style radio :selected (equal (car bhl-i18n-conventions) "en")]
     ["French" (progn (setcar bhl-i18n-conventions "fr") (bhl-mode))
      :style radio :selected (equal (car bhl-i18n-conventions) "fr")]
     ["German" (progn (setcar bhl-i18n-conventions "de") (bhl-mode))
      :style radio :selected (equal (car bhl-i18n-conventions) "de")])
    ["Check dubious tags" bhl-toggle-check-tags
     :style toggle :selected bhl-check-tags-flag]
    ["Numbered sections" bhl-toggle-numbered-section
     :style toggle :selected bhl-numbered-section-flag]
    ["Ask for caption" bhl-toggle-table-ask-caption
     :style toggle :selected bhl-table-ask-caption-flag]
    "---"
    ["Mail buffer" bhl-compose-mail :help "Mail this buffer in plain text."]
    "---"
    ;; miscellaneous (again)
    ["View last conversion log" bhl-view-log t]
    ["Submit bug report" bhl-submit-bug-report t]
    ["Customize BHL" (customize-group "bhl") t]
    ["BHL info manual" (info "bhl.info") :help "The info manual of the BHL mode."]
    ["BHL version" bhl-show-version t]))


;;;; 6 - OTHER FUNCTIONS

;; Follow and go to URL

(defun bhl-goto-next-url ()
  "Go to the next URL."
  (interactive)
  (or (re-search-forward bhl-generic-url-regexp nil t)
      (and (goto-char (point-min))
	   (re-search-forward bhl-url-regexp nil t))))

(defun bhl-goto-previous-url ()
  "Go to the next URL."
  (interactive)
  (or (re-search-backward bhl-generic-url-regexp nil t)
      (and (goto-char (point-max))
	   (re-search-backward bhl-url-regexp nil t))))

(defun bhl-url-at-point ()
  "Return non-nil if a URL is at point."
  (save-excursion
    (skip-chars-backward "^'\"<>[{}( \t\n")
    (looking-at bhl-generic-url-regexp)))

(defun bhl-visit-url (link)
  "Visit the URL named LINK."
  (funcall bhl-browse-url-function link))

(defun bhl-follow-url-at-point ()
  "Visit the link at point."
  (interactive)
  (when (bhl-url-at-point)
    (cond ((equal (substring (match-string 0) 0 7) "mailto:")
	   (compose-mail (substring (match-string 0) 7)))
	  (t (bhl-visit-url (or (match-string 5) (match-string 0)))))))

(defun bhl-follow-url-at-mouse (event)
  "According to EVENT, visit the link at point."
  (interactive "e")
  (save-excursion
    (when (fboundp 'posn-window)
      (set-buffer (window-buffer (posn-window (event-start event))))
      (goto-char (posn-point (event-start event))))
    (when (bhl-url-at-point)
      (cond ((equal (substring (match-string 0) 0 7) "mailto:")
	     (compose-mail (substring (match-string 0) 7)))
	    (t (bhl-visit-url (or (match-string 5) (match-string 0))))))))

;; Toggle main variables

(defun bhl-toggle-check-tags ()
  "Toggle `bhl-check-tags-flag'."
  (interactive)
  (setq bhl-check-tags-flag (not bhl-check-tags-flag))
  (bhl-mode))

(defun bhl-toggle-numbered-section ()
  "Toggle `bhl-numbered-section-flag'."
  (interactive)
  (setq bhl-numbered-section-flag (not bhl-numbered-section-flag))
  (bhl-mode))

(defun bhl-toggle-table-ask-caption ()
  "Toggle `bhl-table-ask-caption-flag'."
  (interactive)
  (setq bhl-table-ask-caption-flag (not bhl-table-ask-caption-flag))
  (bhl-mode))

;; Version and Tab

(defun bhl-show-version ()
  "Display version infos about BHL."
  (interactive)
  (message "BHL mode %s, Copyright (C) 2003 Bastien Guerry; type C-h m for help."
	   bhl-version)
  (if (sit-for 3)
      (let ((lines bhl-startup-message-lines))
	(while (and (sit-for 3) lines)
	  (message (substitute-command-keys (car lines)))
	  (setq lines (cdr lines)))))
  (message ""))

;; Don't INDENT paragraphs. Restrict indentation to lists.

(defun bhl-tab ()
  "Tabulation for the BHL mode.
Use this function to indent list items.
Don't indent paragraphs."
  (interactive)
  (indent-to (+ (current-column) tab-width)))

;; Move inside the BHL source:

(defun bhl-goto-next-section (&optional arg)
  "Go to the title of the next section.
If ARG is non-nil, go to the title of the next section or (sub)subsection."
  (interactive "P")
  (let ((reg (cond (arg "^\\([0-9]+\.\\)+")
		   (t "^[0-9]+. "))))
    (re-search-forward reg nil t)))

(defun bhl-goto-previous-section (&optional arg)
  "Go to the title of the previous section.
If ARG is non-nil, go to the title of the previous section or (sub)subsection."
  (interactive "P")
  (let ((reg (cond (arg "^\\([0-9]+\.\\)+")
		   (t "^[0-9]+\. "))))
    (re-search-backward "^[0-9]+\. " nil t)))

;; Font handling:

(define-skeleton bhl-change-font-bold
  "Change font to bold."
  nil
  "*" _ "*")

(define-skeleton bhl-change-font-emph
  "Change font to emphasis."
  nil
  "_" _ "_")

(define-skeleton bhl-change-font-tt
  "Change font to truetype."
  nil
  "==" _ "==")

(define-skeleton bhl-change-font-under
  "Change font to underline."
  nil
  "__" _ "__")

(define-skeleton bhl-change-font-bolditalic
  "Change font to bold-italic."
  nil
  "_*" _ "*_")

(defun bhl-compose-mail ()
  "Mail the content of the BHL buffer in plain text."
  (interactive)
  (bhl2txt)
  (save-buffer)
  (let ((file (buffer-file-name)))
    (compose-mail)
    (save-excursion
      (goto-char (point-max))
      (setq tab-width 3)
      (insert-file-contents file))))

(defun bhl-check-comment (arg)
  "Check for dubious comment strings in the ARG output."
  (save-restriction
    (save-excursion
      (let ((regexp (cond ((equal arg "latex") "%+")
			  (t "<!--\\|-->"))))
	(goto-char (point-min))
	(while (re-search-forward regexp nil t)
	  (add-to-list 'bhl-conversion-log
		       (format "Dubious comment %s at point %s."
			       (match-string 0)
			       (match-beginning 0) t)))))))

(defun bhl-check-existing-tags (arg)
  "Check if there is no tag of the ARG format in the source file."
  (let ((reg (cond ((equal arg "latex")
		    "\\(\\\\[^ \t\n{]+\\)\\({\\([^}]+\\)}\\)?")
		   (t "\\(</?[^>\t\n]+>\\)"))))
    (save-excursion
      (save-match-data
	(goto-char (point-min))
	(while (re-search-forward reg nil t)
	  (let (msg)
	    (setq msg (format "Dubious %s at point %d: "
			      (match-string 1)
			      (match-beginning 0)))
	    (if (y-or-n-p (format "%s: do you want to clean up? "
				  (match-string 0)))
		(progn (add-to-list 'bhl-conversion-log
				    (concat msg "deleted.") t)
		       (replace-match 
			(or (match-string 3) "")))
	      (add-to-list 'bhl-conversion-log
			   (concat msg "NOT deleted.") t))))))))

;; Highlighting functions

(defun bhl-url-highlight (limit)
  "Highlight URLs in the BHL buffer limited to LIMIT."
  (while (re-search-forward bhl-url-regexp limit t)
    (add-text-properties (match-beginning 0) (match-end 0)
			 '(face bhl-url-face))))

(defun bhl-generic-url-highlight (limit)
  "Highlight URLs in the BHL buffer limited to LIMIT."
  (while (re-search-forward bhl-generic-url-regexp limit t)
    (let ((map (make-sparse-keymap)))
      (define-key map [mouse-2] 'bhl-follow-url-at-mouse)
      (define-key map [(return)] 'bhl-follow-url-at-point)
      (add-text-properties (match-beginning 0) (match-end 0)
			   `(local-map ,map mouse-face highlight help-echo
				       "mouse-2: follow URL")))))

(defun bhl-highlight-toc-buffer ()
  "Highlight the *TOC* buffer."
  (interactive)
  (save-excursion
    (save-match-data
      (goto-char (point-min))
      (setq buffer-read-only nil)
      (while (re-search-forward "[ ]*\\([0-9]+\.\\)+ \\(.*\\)$" nil t)
	(add-text-properties
	 (match-beginning 2) (match-end 2)
	 '(mouse-face highlight help-echo "mouse-2: go to this section"))))))
					  
;; Handle footnotes

(defun bhl-convert-footnotes (arg)
  "Convert footnotes into ARG format."
  (let ((ftn-regexp
	 (concat (regexp-quote footnote-start-tag)
		 (eval (intern
			(concat "footnote-"
				(symbol-name footnote-style)
				"-regexp")))
		 (regexp-quote footnote-end-tag))))
    (goto-char (point-min))
    (when (and (re-search-forward footnote-section-tag-regexp nil t)
	       (equal arg "html"))
      (replace-match (concat "<P><B>" (match-string 0) "</B>")))
    (while (re-search-forward (concat "\\(" ftn-regexp "\\)[ ]*") nil t)
      (let* ((num (match-string 1)))
	(when (equal arg "html")
	  (replace-match (concat
			  "<P><A name=\"" num
			  "\" href=\"#ftn" num "\">"
			  num "</a>&nbsp;&nbsp;")))
	(copy-to-register 255 (point)
			  (progn
			    (re-search-forward "\\[\\|^$" nil t)
			    (goto-char (1- (match-beginning 0)))))
	(funcall 'bhl-replace-footnote num arg)))
    (when (equal arg "latex")
      (bhl-skip-footnotes))
    (add-to-list 'bhl-conversion-log "Footnotes conversion.")))

(defun bhl-skip-footnotes (&optional arg)
  "Skip footnotes according to ARG format."
  (save-excursion
    (goto-char (point-min))
    (when arg
      (while (re-search-forward
	      (concat (regexp-quote footnote-start-tag)
		      (eval (intern
			     (concat "footnote-"
				     (symbol-name footnote-style)
				     "-regexp")))
		      (regexp-quote footnote-end-tag)) nil t)
	(replace-match "")))
    (goto-char (point-min))
    (when (re-search-forward footnote-section-tag-regexp nil t)
      (kill-region (match-beginning 0) (point-max)))))

(defun bhl-replace-footnote (num arg)
  "Replace the footnote number NUM into ARG format."
  (save-excursion
    (goto-char (point-min))
    (re-search-forward (regexp-quote num) nil t)
    (cond ((equal arg "latex")
	   (replace-match (concat "\\\\footnote{" (get-register 255) "}")))
	  ((equal arg "html")
	   (replace-match
	    (concat "<SUP><a name=\"ftn" (match-string 0)
		    "\" href=\"#" (match-string 0) "\">"
		    (match-string 0) "</a></SUP>"))))))

;; Convert verbatim, minipages and horizontal rules

(defun bhl-convert-verbatim (arg)
  "Convert verbatim environment into ARG format."
  (bhl-convert-special arg
		       '("" "<PRE>" "<P><VERB>" "\\begin{verbatim}")
		       '("" "</PRE>" "</VERB>" "\\end{verbatim}")
		       "--" "--"))

(defun bhl-convert-hr (arg)
  "Convert a raw of dashes into an horizontal rule.
ARG is the output format."
  (save-excursion
    (goto-char (point-min))
    (let (justify width)
      (while (re-search-forward bhl-hr-regexp nil t)
	(cond ((equal (match-string 1) "<")
	       (setq justify "left"))
	      ((equal (match-string 1) ">")
	       (setq justify "right"))
	      (t (setq justify "center")))
	(cond ((equal arg "txt")
	       (if (match-string 4)
		   (setq width (string-to-number (match-string 4)))
		 (setq width (or fill-column 70))))
	      ((equal arg "latex")
	       (setq width (or (match-string 3) "\\\\textwidth")))
	      ((equal arg "html")
	       (setq width (or (match-string 3) "100%"))))
	(cond ((equal arg "txt")
	       (let (replacmt)
		 (if (equal justify "right")
		     (setq replacmt (concat (make-string (- (or fill-column 70) width) 32)
					    (make-string width ?-)))
		   (setq replacmt (make-string (or width fill-column 70) ?-)))
		 (replace-match replacmt)
		 (when (equal justify "center")
		   (center-line))))
	      ((equal arg "html")
	       (replace-match (concat "<HR style=\"width: "
				      width ";\" align=\"" justify "\">")))
	      ((equal arg "latex")
	       (let (flush)
		 (setq flush (if (equal justify "center") "" "flush"))
		 (replace-match
		  (concat "\\\\begin{" flush justify "}\n\\\\rule{"
			  width "}{1pt}\n\\\\end{" flush justify"}"))))
	      ((equal arg "sgml")
	       (replace-match "")))))))

(defun bhl-convert-mpage (arg)
  "Convert minipage into ARG format."
  (save-excursion
    (goto-char (point-min))
    (while (re-search-forward bhl-mpage-regexp nil t)
      (let ((border (if (equal (match-string 2) "-")
			(cond ((equal arg "html")
			       " border: 1pt solid;")
			      ((equal arg "latex")
			       "\\fbox{")
			      (t "")) ""))
	    (end-border (when (and (equal arg "latex")
				   (equal (match-string 2) "-")) "}"))
	    (bhl-intro-mpage (regexp-quote (match-string 0)))
	    justify width intro-mpage end-mpage)
	(cond ((equal arg "latex")
	       (cond ((equal (match-string 1) "<")
		      (setq justify "flushleft"))
		     ((equal (match-string 1) ">")
		      (setq justify "flushright"))
		     (t (setq justify "center")))
	       (if (match-string 4) (setq width (match-string 4))
		 (setq width "\\textwidth")))
	      ((equal arg "html")
	       (cond ((equal (match-string 1) "<")
		      (setq justify "left"))
		     ((equal (match-string 1) ">")
		      (setq justify "right"))
		     (t (setq justify "center")))
	       (if (match-string 4) (setq width (match-string 4))
		 (setq width "100%"))))
	(setq intro-mpage
	      `("" ,(concat "<DIV align=\"" justify "\">\n"
			    "<DIV style=\"padding: 2mm; width: "
			    width ";" border "\" align=\"justify\">")
		"<P><CODE>"
		,(concat
		  "\\vspace{.3cm}\n\\begin{" justify
		  "}\n" border
		  "\\begin{minipage}[c]{" width "}")))
	(setq end-mpage `("" "</DIV>\n</DIV>" "</CODE>"
			  ,(concat "\\end{minipage}" end-border
				   "\n\\end{" justify "}\n\\vspace{.3cm}\n")))
	(bhl-convert-special arg intro-mpage end-mpage bhl-intro-mpage "^%%+")))))

(defun bhl-convert-special (arg intro-tag end-tag mark-beg mark-end)
  "Convert a specific region into ARG format.
INTRO-TAG is a list of TXT, HTML, SGML and LaTeX tags to
insert at the beginning of the region.
END-TAG is a list of TXT, HTML, SGML and LaTeX tags  to
insert at the end of the region.
MARK-BEG is the delimiter of the beginning of the region.
MARK-END is the delimiter of the end of the region."
  (save-restriction
    (save-excursion
      (goto-char (point-min))
      (let ((prefix (cond ((equal arg "txt") (nth 0 intro-tag))
			  ((equal arg "html") (nth 1 intro-tag))
			  ((equal arg "sgml") (nth 2 intro-tag))
			  ((equal arg "latex") (nth 3 intro-tag))))
	    (suffix (cond ((equal arg "txt") (nth 0 end-tag))
			  ((equal arg "html") (nth 1 end-tag))
			  ((equal arg "sgml") (nth 2 end-tag))
			  ((equal arg "latex") (nth 3 end-tag)))))
	(when (re-search-forward (concat "^\\(" mark-beg "\\)$") nil t)
	  (replace-match prefix t t nil 1))
	(when (re-search-forward (concat "^\\(" mark-end "\\)$") nil t)
	  (replace-match suffix t t nil 1))))))

;; Insertion functions

(defun bhl-insert-special (arg)
  "Insert a special environment, according to ARG.
When ARG is a dash, insert a horizontal rule.
When ARG is \"%\", insert a minipage."
  (let ((mk "")
	(align (completing-read
		"Align: " '(("left" 1) ("center" 2) ("right" 3))
		nil t "center"))
	(width (read-from-minibuffer "Width: ")))
    (cond ((equal align "left") (setq mk "<"))
	  ((equal align "center") (setq mk arg))
	  ((equal align "right") (setq mk ">")))
    (insert (concat mk (make-string 3 (string-to-char arg))
		    (if (equal width "") "\n" (concat "[" width "]\n"))))))

(defun bhl-insert-url ()
  "Read an URL and its name from the minibuffer."
  (interactive)
  (let ((url (read-from-minibuffer "URL: " ""))
	(name (read-from-minibuffer "Name: ")))
    (if (equal name "")
	(insert url)
      (insert "[[" url "][" name "]]"))
    (message "URL \"%s\" named \"%s\" inserted." url name)))

(defun bhl-insert-img (filename)
  "Read a FILENAME to insert as an image."
  (interactive "FInsert image: ")
  (if (file-directory-p filename)
      (signal 'file-error (list "Opening input file" "file is a directory"
				filename))
    (insert "[[" (file-name-nondirectory filename) "]]")
    (message "Image \"%s\" inserted." (file-name-nondirectory filename))))

(defun bhl-insert-hr ()
  "Insert a horizontal rule."
  (interactive)
  (bhl-insert-special "-")
  (message "Horizontal rule inserted."))

(defun bhl-insert-mpage ()
  "Insert a minipage."
  (interactive)
  (bhl-insert-special "%")
  (save-excursion
    (insert "\n%%%%"))
  (message "Minipage inserted."))

(defun bhl-insert-verbatim ()
  "Insert a verbatim environment."
  (interactive)
  (insert "--\n")
  (save-excursion
    (insert "--"))
  (message "Verbatim environment inserted."))

;; Core of the conversion:

(defun bhl-replace-section (sec beg end &optional acr num ctr)
  "Replace section's labels with the right tags (HTML or LaTeX).
Here are the arguments:

SEC:  a regexp which matches the section;
BEG:  a string to be inserted just before the title of the section;
END:  a string to be inserted just after the title of the section;

ACR (optional): this argument specifies if an anchor is required;
NUM (optional): if nil, this argument prevents the numbering of sections.
  See `bhl-numbered-section-flag' for details.
CTR (optional): if t, center line."
  (goto-char (point-min))
  (while (re-search-forward (concat "\\(" sec "\\)\\(.*$\\)") nil t)
    (let ((str (match-string 0)))
      (replace-match
       (concat beg
	       (if num (match-string 1) "")
	       (match-string 2) end) t t)
      (when ctr (center-line))
      (when acr
	(beginning-of-line)
	(insert "<A name=\"" (bhl-string-to-anchor str) "\"></A>\n"))
      (goto-char (point-min)))))

(defun bhl-replace-tag (word beg end)
  "Replace tags (HTML or LaTeX).
WORD is the string to be replaced.
BEG and END are the beginning end ending tags."
  (goto-char (point-min))
  (while (re-search-forward word nil t)
    (replace-match (concat (regexp-quote beg) (match-string 2) end) 
						 t nil nil 1)
    (goto-char (point-min))))

;; A generic conversion function for punctuation and special characters.
(defun bhl-convert-char (arg lst)
  "Convert into the ARG format some strings from LST."
  (save-excursion
    (let ((cpt 0)
	  (case-fold-search nil)
	  (convert-char-list lst))
      (while (nth cpt convert-char-list)
	(goto-char (point-min))
	(while (re-search-forward (car (nth cpt convert-char-list)) nil t)
	  (replace-match
	   (cond ((equal arg "latex")
		  (concat (caddr (nth cpt convert-char-list))))
		 ((or (equal arg "html") (equal arg "sgml"))
		  (if (and (equal arg "sgml")
			   (or (equal (match-string 0) "oe")
			       (equal (match-string 0) "OE")))
		      (match-string 0)
		    (concat (cadr (nth cpt convert-char-list)))))
		 (t (match-string 0)))
	   t t nil (if (match-string 1) 1 0)))
	(setq cpt (1+ cpt))))))

(defun bhl-convert-tex (arg)
  "Convert \"LaTeX\", \"LaTeX2e\" and \"TeX\" strings into ARG format."
  (save-excursion
    (goto-char (point-min))
    (let ((case-fold-search nil))
      (while (re-search-forward "\\<\\(LaTeX\\)\\(2\\)?\\(e\\)?\\>" nil t)
	(replace-match (cond ((equal arg "latex")
			      (concat "\\LaTeX" (match-string 3) "{}"))
			     ((equal arg "html")
			      (concat "L<SUP>A</SUP>T<SUB>E</SUB>X"
				      (if (match-string 3) (match-string 2))
				      (if (match-string 3) "&#949;")))
			     (t (match-string 0))) t t))
      (goto-char (point-min))
      (while (re-search-forward "\\<TeX\\>" nil t)
	(cond ((equal arg "latex") (replace-match "\\TeX{}" t t))
	      ((equal arg "html") (replace-match "T<SUB>E</SUB>X" t t))
	      (t (match-string 0) t t))))))

(defun bhl-convert-img (arg)
  "Convert inserted images into ARG format."
  (save-excursion
    (goto-char (point-min))
    (cond ((equal arg "txt")
	   (while (re-search-forward bhl-img-regexp nil t)
	     (replace-match "")))
	  ((equal arg "html")
	   (while (re-search-forward bhl-img-regexp nil t)
	     (replace-match
	      (concat "<P><IMG alt=\"" (match-string 1)
		      "\" src=\"" (match-string 1) "\" align=\""
		      bhl-html-img-align "\">"))))
	  ((equal arg "sgml")
	   (while (re-search-forward bhl-img-regexp nil t)
	     (replace-match
	      (concat "<!-- " (match-string 1) " -->"))))
	  ((equal arg "latex")
	   (while (re-search-forward bhl-img-regexp nil t)
	     (replace-match
	      (concat "% " (match-string 1)))))
	  (t (while (re-search-forward bhl-img-regexp nil t)
	       (replace-match ""))))))

(defun bhl-convert-list-intern (reg open item close)
  "Convert lists.
REG is the regular expression to search for.
OPEN is the opening tag of the list.
ITEM is the tag for each item of the list.
CLOSE is the closing tag of the list."
  (save-restriction
    (save-excursion
      (save-match-data
	(let ((beg (progn (re-search-forward reg nil t)
			  (match-beginning 0)))
	      (sep (match-string 0))
	      (end (progn (re-search-forward "^[^ \t\n]\\|\\'" nil t)
			  (match-beginning 0))))
	  (narrow-to-region beg end)
	  (goto-char (point-min))
	  (insert open "\n")
	  (while (re-search-forward reg nil t)
	    (replace-match item))
	  (re-search-forward "^$" nil t)
	  (insert close "\n")
	  (widen))))))

(defun bhl-convert-list-regexp ()
  "Define which regexp is converted into a list."
  (cond ((match-string 2) "^\t\t\t\\([-*o]\\|[0-9]+\.\\) ")
	((match-string 1) "^\t\t\\([-*o]\\|[0-9]+\.\\) ")
	(t "^\t\\([-*o]\\|[0-9]+\.\\) ")))

(defun bhl-convert-list (arg)
  "Ordered and non-ordered list converter.
The argument ARG specifies the output format."
  (goto-char (point-min))
  (while (re-search-forward bhl-list-regexp nil t)
    (beginning-of-line)
    (if (or (equal (substring
		    (match-string 0)
		    (- (length (match-string 0)) 2)) "- ")
	    (equal (substring
		    (match-string 0)
		    (- (length (match-string 0)) 2)) "o ")
	    (equal (substring
		    (match-string 0)
		    (- (length (match-string 0)) 2)) "* "))
	(cond ((equal arg "html")
	       (bhl-convert-list-intern
		(bhl-convert-list-regexp)
		"<UL>" "<LI>" "</UL>"))
	      ((equal arg "sgml")
	       (bhl-convert-list-intern
		(bhl-convert-list-regexp)
		"<P><ITEMIZE>" "<ITEM>" "</ITEMIZE>"))
	      ((equal arg "latex")
	       (bhl-convert-list-intern
		(bhl-convert-list-regexp)
		"\\begin{itemize}" "\\\\item " "\\end{itemize}")))
      (cond ((equal arg "html")
	     (bhl-convert-list-intern
	      (bhl-convert-list-regexp)
	      "<OL>" "<LI>" "</OL>"))
	    ((equal arg "sgml")
	     (bhl-convert-list-intern
	      (bhl-convert-list-regexp)
	      "<P><ENUM>" "<ITEM>" "</ENUM>"))
	    ((equal arg "latex")
	     (bhl-convert-list-intern
	      (bhl-convert-list-regexp)
	      "\\begin{enumerate}" "\\\\item " "\\end{enumerate}"))))
    (goto-char (point-min))))

(defun bhl-convert-description (arg)
  "Convert list of descriptions.
The argument ARG specifies the output format."
  (goto-char (point-min))
  (re-search-forward "<BODY" nil t)
  (save-restriction
    (let ((beg (point))
	  (end (point-max)))
      (narrow-to-region beg end)
      (while (re-search-forward bhl-descrip-regexp nil t)
	(save-excursion
	  (save-match-data
	    (beginning-of-line)
	    (let ((beg (point))
		  (end (progn
			 (re-search-forward "\\(^[-<[:alnum:]]\\|\\'\\)" nil t)
			 (forward-line -1)
			 (point))))
	      (narrow-to-region beg end)
	      (cond ((equal arg "html")
		     (goto-char (point-min))
		     (insert "<DL>\n")
		     (while (re-search-forward bhl-descrip-regexp nil t)
		       (replace-match
			(concat "<DT>" (match-string 3) "</DT><DD>") t nil nil 2))
		     (goto-char (point-max))
		     (insert "</DL>\n\n"))
		    ((equal arg "sgml")
		     (goto-char (point-min))
		     (insert "<P><DESCRIP>\n")
		     (while (re-search-forward bhl-descrip-regexp nil t)
		       (replace-match
			(concat "<TAG>" (match-string 3) "</TAG> ") t nil nil 2))
		     (goto-char (point-max))
		     (insert "</DESCRIP></P>\n"))
		    ((equal arg "latex")
		     (goto-char (point-min))
		     (insert "\\begin{description}\n")
		     (while (re-search-forward bhl-descrip-regexp nil t)
		       (replace-match
			(concat "\\\\item [" (match-string 3) "] ") t nil nil 2))
		     (goto-char (point-max))
		     (insert "\\end{description}\n")))
	      (widen))))))))

(defun bhl-convert-url (&optional arg)
  "Convert an URL.  An optional argument ARG specifies the output format."
  (save-excursion
    (save-restriction
      (goto-char (point-min))
      (while (re-search-forward bhl-url-regexp nil t)
	(cond
	 ((equal arg "txt")
	  (cond ((equal (match-string 3) "mailto:")
		 (replace-match (match-string 4) nil t))
		((equal (match-string 6) "mailto:")
		 (replace-match (concat (match-string 7) " (" (match-string 8) ")") nil t))
		((match-string 8)
		 (replace-match (concat (match-string 5) " (" (match-string 8) ")") nil t))
		(t (replace-match (match-string 1) nil t))))
	 ((equal arg "html")
	  (if (and (equal (match-string 3) "mailto:")
		   (not (match-string 8)))
	      (replace-match
	       (concat "<A href=\""
		       (match-string 1) "\">"
		       (match-string 4) "</A>") nil t)
	    (replace-match
	     (concat "<A href=\""
		     (or (match-string 5)
			 (match-string 1)) "\">"
		     (or (match-string 8)
			 (match-string 1))"</A>") nil t)))
	 ((equal arg "sgml")
	  (replace-match
	   (concat "<HTMLURL url=\"" (or (match-string 5)
					 (match-string 1))
		   "\" name=\"" (or (match-string 8)
				    (match-string 1)) "\">")))
	 ((equal arg "latex")
	  (cond ((equal (match-string 3) "mailto:")
		 (replace-match
		  (concat "\\url{" (match-string 4) "}") nil t))
		((equal (match-string 6) "mailto:")
		 (replace-match
		  (concat (match-string 8) " (\\url{" (match-string 7) "})") nil t))
		((match-string 8)
		 (replace-match
		  (concat (match-string 8) " (\\url{" (match-string 5) "})") nil t))
		(t (replace-match (concat "\\url{" (match-string 1) "}") nil t)))))))))

(defun bhl2html-convert-table ()
  "Find tables and convert them into HTML format."
  (goto-char (point-min))
  (save-restriction
    (while (re-search-forward "^|" nil t)
      (beginning-of-line)
      (let ((align (cond ((equal bhl-table-cell-align "l") "align=\"left\"")
			 ((equal bhl-table-cell-align "c") "align=\"center\"")
			 ((equal bhl-table-cell-align "r") "align=\"right\"")))
	    (start (point))
	    (end (progn (re-search-forward "^[^ |\t\n]\\|\\'" nil t)
			(beginning-of-line)
			(point))))
	(narrow-to-region start end)
	(goto-char (point-min))
	(let ((caption (if bhl-table-ask-caption-flag
			   (read-from-minibuffer "Caption: ") "No summary" )))
	  (insert "<TABLE align=\"" bhl-table-align "\" summary=\"" caption "\" ")
	  (let ((cpt 0))
	    (while (nth cpt bhl-html-table-param-alist)
	      (insert (concat (car (nth cpt bhl-html-table-param-alist))
			      "=\"" (cdr (nth cpt bhl-html-table-param-alist)) "\" "))
	      (setq cpt (1+ cpt))))
	  (insert ">\n")
	  (when bhl-table-ask-caption-flag
	    (insert "<CAPTION>" caption "</CAPTION>\n")))
	(while (re-search-forward "|\\([^|\n]+\\)|"
				  (save-excursion (end-of-line) (point)) t)
	  (beginning-of-line)
	  (insert "<TR>")
	  (re-search-forward "[ \t]*|" nil t)
	  (replace-match "")
	  (while (re-search-forward
		  "\\([ \t]*\\)\\(\*?\\)\\([^|*\n\t]+\\)\\(\*?\\)\\([ \t]*\\)|"
		  (save-excursion (end-of-line) (point)) t)
	    (replace-match
	     (concat (if (equal (match-string 2) "*")
			 (concat "<TH " align ">") (concat "<TD " align ">"))
		     (match-string 3)
		     (if (equal (match-string 2) "*") "</TH>" "</TD>"))))
	  (insert "</TR>")
	  (forward-line 1))
	(insert "</TABLE>\n")
	(widen)))
    (goto-char (point-min))))

(defun bhl2xxx-convert-table (arg)
  "Find tables and convert them into ARG format."
  (goto-char (point-min))
  (save-restriction
    (while (re-search-forward "^|" nil t)
      (beginning-of-line)
      (let* ((count 0)
	     (latex (equal arg "latex"))
	     (table-header "")
	     (make-header nil)
	     (start (point))
	     (end (progn (re-search-forward "^[^ |\t\n]\\|\\'" nil t)
			 (beginning-of-line)
			 (point))))
	(narrow-to-region start end)
	(goto-char (point-min))
	(while (re-search-forward "|\\([^|\n]+\\)|"
				  (save-excursion (end-of-line) (point)) t)
	  (setq count (1+ count))
	  (beginning-of-line)
	  (while (re-search-forward
		  "|\\([ \t]*\\)\\(\*?\\)\\([^|*\n\t]+\\)\\(\*?\\)\\([ \t]*\\)"
		  (save-excursion (end-of-line) (point)) t)
	    (when (equal count 1)
	      (progn (setq table-header (concat table-header bhl-table-cell-align))
		     (if (equal (match-string 2) "*")
			 (setq make-header t)
		       (setq make-header nil))))
	    (replace-match
	     (concat (match-string 3) (if latex "&" "|"))))
	  (forward-char -1)
	  (kill-region (point) (progn (end-of-line) (point)))
	  (insert (if latex "\\\\" "@"))
	  (when make-header (insert (if latex " \\hline" "<HLINE>")))
	  (forward-line 1)
	  (setq make-header nil))
	(insert (if latex "\\end{tabular}\n" "</TABULAR>\n"))
	(if latex
	    (when bhl-table-ask-caption-flag
	      (insert (concat  "\\caption{"
			       (read-from-minibuffer "Caption: ") "}\n")))
	  (when bhl-table-ask-caption-flag
	    (insert (concat  "<CAPTION>"
			     (read-from-minibuffer "Caption: ")
			     "</CAPTION>\n"))))
	(insert (if latex "\\end{table}\n" "</TABLE>\n"))
	(setq table-header
	      (if latex
		  (concat
		   "\\begin{table}[" bhl-table-location
		   "]\n"
		   (cond ((equal bhl-table-align "left") "\\flushleft")
			 ((equal bhl-table-align "center") "\\centering")
			 ((equal bhl-table-align "right") "\\flushright"))
		   "\n\\begin{tabular}{" table-header "}\n")
		(concat "<P><TABLE loc=\"" bhl-table-location
			"\">\n<TABULAR ca=\"" table-header "\">\n")))
	(goto-char start)
	(insert table-header)
	(goto-char (point-max))
	(widen)))
    (goto-char (point-min))))

(defun bhl-insert-preamble (arg)
  "Insert the header of `bhl2html', `bhl2latex' and `bhl2sgml'.
ARG defines the output format."
  (goto-char (point-min))
  (let ((bhl-title (save-excursion (re-search-forward bhl-title-regexp nil t)
				   (match-string 2))))
    (cond ((equal arg "html")
	   (insert bhl-html-doctype "\n")
	   (insert "<HTML>\n<HEAD>\n" bhl-html-content-type "\n")
	   (insert "<LINK href=\"http://bhl.mode.free.fr\" rel=\"generator-home\">\n")
	   (let ((cpt 0))
	     (while (< cpt (safe-length bhl-html-meta-alist))
	       (insert "<META name=\"" (car (nth cpt bhl-html-meta-alist))
		       "\" content=\"" (cdr (nth cpt bhl-html-meta-alist)) "\">\n")
	       (setq cpt (1+ cpt))))
	   (insert "<TITLE>" bhl-title "</TITLE>\n")
	   (if bhl-html-style
	       (insert bhl-html-style "\n"))
	   (insert "</HEAD>\n\n")
	   (insert "<BODY>\n")
	   (search-forward bhl-title nil t)
	   (replace-match (concat (car bhl-html-title-tags)
				  bhl-title (cadr bhl-html-title-tags))))
	  ((equal arg "sgml")
	   (replace-match "")
	   (insert "<!DOCTYPE linuxdoc SYSTEM>\n<ARTICLE>")
	   (insert "<TITLE>" bhl-title "</TITLE>\n")
	   (insert "<AUTHOR>" bhl-author-name "</AUTHOR>\n\n")
	   (insert "<ABSTRACT>")
	   (save-excursion
	     (re-search-forward "^[0-9]+\." nil t)
	     (forward-line -1)
	     (insert "</ABSTRACT>\n\n")))
	  ((equal arg "latex")
	   (replace-match "")
	   (insert "\\documentclass[" (number-to-string bhl-latex-pt)
		   "pt]{" bhl-latex-class "}\n")
	   (let ((cpt 0))
	     (while (< cpt (safe-length bhl-latex-packages-alist))
	       (insert "\\usepackage")
	       (unless (equal "" (cdr (nth cpt bhl-latex-packages-alist)))
		 (insert "[" (cdr (nth cpt bhl-latex-packages-alist)) "]"))
	       (insert "{" (car (nth cpt bhl-latex-packages-alist)) "}\n")
	       (setq cpt (1+ cpt))))
	   (insert "\n\\title{" bhl-title "}\n\\author{"
		   bhl-author-name "}\n\n\\begin{document}\n")
	   (insert "\\maketitle\n"))
	  (t (insert "%% ") (end-of-line) (insert " %%")
	     (when bhl-txt-sec-center-flag (center-line))))))


;;;; 7 - TABLE OF CONTENTS

(defun bhl-update-toc ()
  "Update the table of contents.
Update the list of section via `bhl-tsl'
and the list of corresponding points via `bhl-tpl'."
  (interactive)
  (setq bhl-tsl nil)
  (setq bhl-tpl nil)
  (save-restriction
    (let ((position (point))
	  (scpt 0))
      (goto-char (point-min))
      (while (re-search-forward "\\(^[0-9]+\.\\)\\( .*$\\)" nil t)
	(setq scpt (1+ scpt))
	(save-match-data
	  (replace-match (concat (number-to-string scpt)
				 "." (match-string 2))))
	(save-excursion
	  (let ((section (match-string 0))
		(section-point (match-end 0))
		(beg (point))
		(end (progn (re-search-forward "\\(^[0-9]+\. .*$\\|\\'\\)" nil t)
			    (beginning-of-line)
			    (point))))
	    (narrow-to-region beg end)
	    (goto-char (point-min))
	    (let ((sscpt 0) (ss-list nil) (pp-list nil))
	      (while (re-search-forward "\\(^\\([0-9]+\.\\)\\{2\\}\\)\\( .*$\\)" nil t)
		(setq sscpt (1+ sscpt))
		(save-match-data
		  (replace-match (concat (number-to-string scpt) "."
					 (number-to-string sscpt) "." (match-string 3))))
		(let ((ssection (match-string 0))
		      (ssection-point (match-end 0))
		      (beg (point))
		      (end (progn
			     (re-search-forward
			      "\\(^\\([0-9]+\.\\)\\{2\\} .*$\\|^[0-9]+\. .*$\\|\\'\\)" nil t)
			     (beginning-of-line)
			     (point))))
		  (narrow-to-region beg end)
		  (goto-char (point-min))
		  (let ((ssscpt 0) (sss-list nil) (ppp-list nil))
		    (while (re-search-forward "\\(^\\([0-9]+\.\\)\\{3\\}\\)\\( .*$\\)" nil t)
		      (setq ssscpt (1+ ssscpt))
		      (let ((sssection (concat (number-to-string scpt) "."
					       (number-to-string sscpt) "."
					       (number-to-string ssscpt) "."
					       (match-string 3)))
			    (sssection-point (match-end 3)))
			(replace-match sssection)
			(add-to-list 'sss-list sssection t)
			(add-to-list 'ppp-list sssection-point t)))
		    (add-to-list 'ss-list (cons ssection (list sss-list)) t)
		    (add-to-list 'pp-list (cons ssection-point (list ppp-list)) t)))
		(narrow-to-region beg end))
	      (add-to-list 'bhl-tsl (cons section (list ss-list)) t)
	      (add-to-list 'bhl-tpl (cons section-point (list pp-list)) t))
	    (widen))))
      (goto-char position))))

(defun bhl-view-log ()
  "Display the log messages of the last ARG conversion."
  (interactive)
  (with-output-to-temp-buffer "*Log*"
    (princ "BHL conversion log:\n")
    (princ "===================\n\n")
    (let ((cpt 0))
      (when (equal bhl-conversion-log nil)
	(princ "* Last conversion correctly achieved."))
      (while (nth cpt bhl-conversion-log)
	(princ (concat "* " (nth cpt bhl-conversion-log) "\n"))
	(setq cpt (1+ cpt))))))

(defun bhl-string-to-anchor (str)
  "Convert the string STR into an anchor (shorter, without whitespace)."
  (substring str 0 (if (< (length str) 10) (length str) 10)))
    
;; Ok, the next function is quite confused.
(defun bhl-generate-toc (arg &optional txt)
  "Generate the table of contents.
The mandatory argument ARG specifies the toc depth (1-3).
The optional argument TXT:
   \"html\" means \"insert toc into HTML output\".
   t means \"insert toc here\"."
  (setq bhl-point-list nil)
  (let ((cpt 0))
    (while (< cpt (length bhl-tsl))
      (when (and (equal txt "html") (equal cpt 0)
		 (> (length bhl-tsl) 0))
	(insert "<UL>\n"))
      (cond ((equal txt "html")
	     (insert "<LI><A href=\"#")
	     (insert (bhl-string-to-anchor (car (nth cpt bhl-tsl))))
	     (insert "\">" (car (nth cpt bhl-tsl)) "</A></LI>\n"))
	    ((eval-expression txt)
	     (insert " " (car (nth cpt bhl-tsl))"\n" ))
	    (t (add-to-list 'bhl-point-list (car (nth cpt bhl-tpl)) t)
	       (insert " " (car (nth cpt bhl-tsl)) "\n")))
      (when (or (not arg) (> arg 1))
	(let ((cpt2 0))
	  (while (< cpt2 (length (nth 1 (nth cpt bhl-tsl))))
	    (when (and (equal txt "html")
		       (equal cpt2 0)
		       (> (length (nth 1 (nth cpt bhl-tsl))) 0))
	      (insert "   <UL>\n"))
	    (cond ((equal txt "html")
		   (insert "   <LI><A href=\"#")
		   (insert (bhl-string-to-anchor (car (nth cpt2 (nth 1 (nth cpt bhl-tsl))))))
		   (insert "\">" (car (nth cpt2 (nth 1 (nth cpt bhl-tsl)))) "</A></LI>\n"))
		  ((eval-expression txt)
		   (insert "    " (car (nth cpt2 (nth 1 (nth cpt bhl-tsl)))) "\n"))
		  (t (add-to-list 'bhl-point-list (car (nth cpt2 (nth 1 (nth cpt bhl-tpl)))) t)
		     (insert "    " (car (nth cpt2 (nth 1 (nth cpt bhl-tsl)))) "\n")))
	    (when (or (not arg) (> arg 2))
	      (let ((cpt3 0))
		(while (< cpt3 (length (nth 1 (nth cpt2 (nth 1 (nth cpt bhl-tsl))))))
		  (if (and (equal txt "html")
			   (equal cpt3 0)
			   (> (length (nth 1 (nth cpt2 (nth 1 (nth cpt bhl-tsl))))) 0))
		      (insert "      <UL>\n"))
		  (cond ((equal txt "html")
			 (insert "      <LI><A href=\"#")
			 (insert (bhl-string-to-anchor
				  (nth cpt3 (nth 1 (nth cpt2 (nth 1 (nth cpt bhl-tsl)))))))
			 (insert "\">"
				 (nth cpt3 (nth 1 (nth cpt2 (nth 1 (nth cpt bhl-tsl)))))
				 "</A></LI>\n"))
			((eval-expression txt)
			 (insert "       "
				 (nth cpt3 (nth 1 (nth cpt2 (nth 1 (nth cpt bhl-tsl))))) "\n"))
			(t (add-to-list
			    'bhl-point-list
			    (nth cpt3 (nth 1 (nth cpt2 (nth 1 (nth cpt bhl-tpl))))) t)
			   (insert "        "
				   (nth cpt3 (nth 1 (nth cpt2 (nth 1 (nth cpt bhl-tsl)))))
				   "\n")))
		  (setq cpt3 (1+ cpt3)) )
		(when (and (equal txt "html")
			   (> (length (nth 1 (nth cpt2 (nth 1 (nth cpt bhl-tsl))))) 0))
		  (insert "</UL>"))))
	    (setq cpt2 (1+ cpt2)))
	  (when (and (equal txt "html")
		     (> (length (nth 1 (nth cpt bhl-tsl))) 0)) (insert "</UL>"))))
      (setq cpt (1+ cpt)) )
    (when (and (equal txt "html") (> (length bhl-tsl) 0)) (insert "</UL>\n"))))

(defun bhl-show-toc (&optional depth)
  "Open a new buffer named *TOC* and display the browsable table of contents.
The optional argument DEPTH specifies the toc depth."
  (interactive "P")
  (bhl-update-toc)
  (switch-to-buffer-other-window (get-buffer-create "*TOC*"))
  (setq buffer-read-only nil)
  (erase-buffer)
  (insert (propertize bhl-intro-toc 'face '(t bhl-toc-face))
	  " (? for quick help)\n\n")
  (if (not depth)
      (bhl-generate-toc bhl-default-toc-depth nil)
    (bhl-generate-toc depth nil))
  (bhl-highlight-toc-buffer)
  (setq buffer-read-only t)
  (or (< (- (- (window-height)
	       (count-lines (point-min)  (point-max))) 2) 0)
      (shrink-window
       (- (- (window-height)
	     (count-lines (point-min)  (point-max))) 2)))
  (goto-char (point-min))
  (local-set-key [mouse-2] 'bhl-mouse-toc-visit)
  (local-set-key [?\r] 'bhl-toc-visit-location)
  (local-set-key [left] 'bhl-toc-visit-location-previous)
  (local-set-key [right] 'bhl-toc-visit-location-next)
  (local-set-key [up] 'bhl-toc-visit-location-previous)
  (local-set-key [down] 'bhl-toc-visit-location-next)
  (local-set-key "q" 'bhl-toc-quit)
  (local-set-key "f" 'bhl-toc-quit)
  (local-set-key "v" 'bhl-toc-quit)
  (local-set-key "n" 'bhl-toc-visit-location-next)
  (local-set-key "p" 'bhl-toc-visit-location-previous)
  (local-set-key "<" 'bhl-toc-visit-location-top)
  (local-set-key ">" 'bhl-toc-visit-location-bottom)
  (local-set-key "?" 'bhl-toc-show-help)
  (local-set-key "1" 'bhl-show-toc-1)
  (local-set-key "2" 'bhl-show-toc-2)
  (local-set-key "3" 'bhl-show-toc-3)
  (local-set-key "i" 'bhl-toc-insert-toc))

(defun bhl-mouse-toc-visit (event)
  "Visit the item that is clicked on with EVENT."
  (interactive "e")
  (mouse-set-point event)
  (bhl-toc-visit-location (event-end event)))

(defun bhl-toc-visit-location (arg)
  "In the buffer *TOC*, visit the toc item of the current line.
ARG is the toc location."
  (interactive "P")
  (cond ((eq arg 'top) (goto-line 3))
	((eq arg 'bottom)
	 (goto-char (point-max))
	 (forward-line -1))
	((numberp arg) (forward-line arg)))
  (beginning-of-line)
  (let ((nombre
	 (nth (- (string-to-number
		  (substring (what-line) 5)) 3) bhl-point-list)))
    (when (numberp nombre)
      (switch-to-buffer-other-window bhl-main-buffer)
      (goto-char nombre)
      (recenter 0)
      (switch-to-buffer-other-window "*TOC*"))))

(defun bhl-skip-toc (&optional arg)
  "Convert the TOC according to the ARG output format.
With no ARG, erase the toc in the source file."
  (save-match-data
    (save-excursion
      (goto-char (point-min))
      (let ((rep (point))
	    (beg (progn
		   (search-forward bhl-intro-toc nil t)
		   (progn (beginning-of-line) (point))))
	    (end (progn
		   (end-of-line)
		   (search-forward bhl-end-toc nil t)
		   (forward-line 1)
		   (point))))
	(if (not (equal rep beg))
	    (progn
	      (kill-region beg end)
	      (cond ((equal arg "html")
		     (bhl-generate-toc bhl-default-toc-depth "html"))
		    ((and (equal arg "latex")
			  bhl-numbered-section-flag)
		     (insert "\\setcounter{tocdepth}{"
			     (number-to-string bhl-default-toc-depth)
			     "}\n"  "\\tableofcontents\n")))))))))

(defun bhl-insert-toc (&optional depth)
  "Insert the table of contents in the source file.
Optional argument DEPTH is the toc depth.
TOC place is defined in `bhl-toc-location'."
  (interactive "P")
  (save-excursion (bhl-update-toc))
  (cond ((equal bhl-toc-location "top")
	 (save-excursion
	   (goto-char (point-min))
	   (re-search-forward
	    "\\(\\`[ \t\n\r]*\\)\\([[:alnum:] ]+\\)$" nil t)
	   (forward-line 1)
	   (bhl-skip-toc)
	   (forward-line 1)
	   (insert bhl-intro-toc "\n\n")
	   (bhl-generate-toc (or depth bhl-default-toc-depth) t)
	   (insert "\n" bhl-end-toc "\n")))
	((equal bhl-toc-location "bottom")
	 (save-excursion
	   (bhl-skip-toc)
	   (goto-char (point-max))
	   (insert bhl-intro-toc "\n\n")
	   (bhl-generate-toc (or depth bhl-default-toc-depth) t)
	   (insert "\n" bhl-end-toc)))
	((equal bhl-toc-location "point")
	 (save-excursion
	   (bhl-skip-toc)
	   (insert bhl-intro-toc "\n\n")
	   (bhl-generate-toc (or depth bhl-default-toc-depth) t)
	   (insert "\n" bhl-end-toc))))
  (message (concat "TOC inserted at the " bhl-toc-location ".")))

(defun bhl-toc-insert-toc ()
  "Insert the toc from the *TOC* buffer."
  (interactive)
  (bhl-toc-quit)
  (bhl-insert-toc))

(defun bhl-show-toc-1 ()
  "Display the toc in the *TOC* buffer, DEPTH set to one."
  (interactive)
  (bhl-toc-quit)
  (bhl-show-toc 1))

(defun bhl-show-toc-2 ()
  "Display the toc in the *TOC* buffer, DEPTH set to two."
  (interactive)
  (bhl-toc-quit)
  (bhl-show-toc 2))

(defun bhl-show-toc-3 ()
  "Display the toc in the *TOC* buffer, DEPTH set to three."
  (interactive)
  (bhl-toc-quit)
  (bhl-show-toc 3))

(defun bhl-toc-show-help ()
  "Display a quick help for the browsable toc."
  (interactive)
  (message "[n] - next | [p] - previous | [<] - begin | [>] - end | [123] - depth | [i] insert TOC"))

(defun bhl-toc-visit-location-next ()
  "Visit the next item of the toc."
  (interactive)
  (bhl-toc-visit-location 1))

(defun bhl-toc-visit-location-previous ()
  "Visit the previous item of the toc."
  (interactive)
  (bhl-toc-visit-location -1))

(defun bhl-toc-visit-location-top ()
  "Visit the first item of the toc."
  (interactive)
  (bhl-toc-visit-location 'top))

(defun bhl-toc-visit-location-bottom ()
  "Visit the last item of the toc."
  (interactive)
  (bhl-toc-visit-location 'bottom))

(defun bhl-toc-quit ()
  "Quit the toc."
  (interactive)
  (kill-this-buffer)
  (switch-to-buffer-other-window bhl-main-buffer)
  (delete-other-windows))


;;;; 9 - Submit bug report

(defun bhl-submit-bug-report ()
  "Submit via mail a bug report on BHL Mode."
  (interactive)
  (and
   (y-or-n-p "Do you want to submit a report on BHL Mode? ")
   (require 'reporter)
   (reporter-submit-bug-report
    bhl-help-address
    (concat "BHL Mode " bhl-version)
    (list
     ;; report all important variables
     'bhl-mode-hook
     'bhl-i18n-conventions
     'bhl-numbered-section-flag
     'bhl-check-tags-flag
     'bhl-table-ask-caption-flag
     'bhl-default-toc-depth
     'bhl-toc-location
     'bhl-table-location
     'bhl-table-align
     'bhl-table-cell-align
     'bhl-html-table-param-alist
     'bhl-html-para-align
     'bhl-html-img-align
     'bhl-html-text-color
     'bhl-html-bg-color
     'bhl-html-meta-alist
     'bhl-latex-packages-alist
     'bhl-latex-class
     'bhl-latex-pt
     'bhl-html-conversions-list
     'bhl-latex-conversions-list
     'bhl-sgml-conversions-list
     'bhl-txt-conversions-list
     )
    nil
    nil
    "Hi Bastien,\n\n")))


;;;; 8 - BHL2XXX

(defun bhl2html ()
  "Convert the BHL source file into a HTML file."
  (interactive)
  (bhl2xxx "html"))

(defun bhl2sgml ()
  "Convert the BHL source file into a SGML (Linuxdoc) file."
  (interactive)
  (bhl2xxx "sgml"))

(defun bhl2latex ()
  "Convert the BHL source file into a LaTeX file."
  (interactive)
  (bhl2xxx "latex"))

(defun bhl2txt ()
  "Convert the BHL source file into a TXT file.
Strip font tags and URL/IMG square brackets."
  (interactive)
  (bhl2xxx "txt"))

(defun bhl-convert-init (arg)
  "Initial conversion into ARG format."
  (setq bhl-conversion-log nil)
  (if (equal arg "latex")
      (if bhl-numbered-section-flag
	  (setq bhl-latex-syntax-alist bhl-latex-syntax-num-alist)
	(setq bhl-latex-syntax-alist bhl-latex-syntax-nonum-alist)))
  (setq bhl-xxx-syntax-alist
	(eval-expression (intern (concat "bhl-" arg "-syntax-alist"))))
  (setq bhl-xxx-conversions-list
	(eval-expression (intern (concat "bhl-" arg "-conversions-list"))))
  (let ((bhl-source-file buffer-file-name)
	(bhl-workfile-name
	 (concat (file-name-sans-extension buffer-file-name)
		 (cond ((equal arg "latex") ".tex")
		       (t (concat "." arg))))))
    (find-file-other-window bhl-workfile-name)
    (kill-region (point-min) (point-max))
    (setq tab-width 3)
    (insert-file-contents bhl-source-file)))

(defun bhl-convert-misc-1 (arg)
  "Miscellaneous conversions into ARG format.
This conversions are: special char, verbatim, horizontal rules
and LaTeX/TeX labels."
  (unless (not (nth 8 bhl-xxx-conversions-list))
    (when (nth 1 bhl-i18n-conventions)
      (bhl-convert-char arg
			(eval-expression
			 (intern
			  (concat "bhl-"
				  (car bhl-i18n-conventions)
				  "-punctuation")))))
    (when (nth 2 bhl-i18n-conventions)
      (bhl-convert-char arg
			(eval-expression
			 (intern (concat "bhl-" (car bhl-i18n-conventions)
					 "-quotation-marks"))))
      (when (and (equal (concat "bhl-" (car bhl-i18n-conventions)
				"-quotation-marks")
			"bhl-fr-quotation-marks")
		 (equal arg "latex"))
	(add-to-list 'bhl-conversion-log
		     "French quotation marks require the [T1]{fontenc} package." t)))
    (when (nth 3 bhl-i18n-conventions)
      (bhl-convert-char arg
			(eval-expression
			 (intern (concat "bhl-" (car bhl-i18n-conventions)
					 "-special-chars")))))
    (goto-char (point-min))
    (add-to-list 'bhl-conversion-log
		 (concat "I18n conventions: " (car bhl-i18n-conventions) ".")))
  (unless (not (nth 0 bhl-xxx-conversions-list))
    (bhl-convert-verbatim arg))
  (unless (not (nth 1 bhl-xxx-conversions-list))
    (bhl-convert-hr arg))
  (unless (not (nth 7 bhl-xxx-conversions-list))
    (bhl-convert-tex arg)
    (add-to-list 'bhl-conversion-log
		 "LaTeX and TeX labels conversion."  "."))
  (unless (not (nth 9 bhl-xxx-conversions-list))
    (condition-case nil 
	(progn (require 'footnote)
	       (if (equal arg "sgml")
		   (bhl-skip-footnotes arg)
		 (bhl-convert-footnotes arg)))
      (error (add-to-list 'bhl-conversion-log 
			  "BHL was unable to convert footnotes."))))
  (unless (not (nth 10 bhl-xxx-conversions-list))
    (bhl-convert-mpage arg)))

(defun bhl-convert-misc-2 (arg)
  "Miscellaneous conversions into ARG format.
This conversion are: images, URLs, description lists, lists, tables."
  (unless (not (nth 2 bhl-xxx-conversions-list))
    (bhl-convert-img arg))
  (unless (not (nth 3 bhl-xxx-conversions-list))
    (bhl-convert-url arg)
    (when (equal arg "latex")
      (add-to-list 'bhl-conversion-log
		   "Please check that you have included the URL package." t)))
  (unless (equal arg "txt")
    (unless (not (nth 4 bhl-xxx-conversions-list))
      (bhl-convert-description arg))
    (unless (not (nth 5 bhl-xxx-conversions-list))
      (bhl-convert-list arg))
    (unless (not (nth 6 bhl-xxx-conversions-list))
      (when (not bhl-table-ask-caption-flag)
	(add-to-list 'bhl-conversion-log
		     "No caption was asked for tables." t))
      (if (equal arg "html")
	  (bhl2html-convert-table)
	(bhl2xxx-convert-table arg)))))

(defun bhl-convert-para (arg)
  "Convert paragraphs into ARG format."
  (save-excursion
    (goto-char (point-min))
    (when (equal arg "sgml")
      (re-search-forward "</ABSTRACT>" nil t)
      (end-of-line))
    (while (re-search-forward "^[\\'\"&*_=a-zA-Z0-9]" nil t)
      (goto-char (match-beginning 0))
      (if (not (or (save-excursion (forward-line -1)
				   (looking-at "<\\|[a-zA-Z0-9]"))
		   (looking-at "\\([0-9]+\.\\)")))
	  (save-match-data
	    (cond ((or (equal arg "sgml")
		       (equal bhl-html-para-align "none"))
		   (insert "<P>\n"))
		  (t (insert "\n<P align=\"" 
			     bhl-html-para-align "\">")))
	    (re-search-forward "^[ \t]*$" nil t)
	    (insert "</P>\n"))
	(forward-line 1) ))))

(defun bhl-convert-sec-tag (arg)
  "Convert sections and tags into ARG format."
  (let ((cpt 0))
    (while (< cpt (length bhl-xxx-syntax-alist))
      (let ((mt (nth cpt bhl-syntax-list))
	    (db (car (nth cpt bhl-xxx-syntax-alist)))
	    (fn (cadr (nth cpt bhl-xxx-syntax-alist))))
	(if (< cpt 4)
	    (bhl-replace-tag mt db fn)
	  (bhl-replace-section mt db fn (equal arg "html")
			       (when (equal arg "html")
				 bhl-numbered-section-flag)
			       (when (equal arg "txt")
				 bhl-txt-sec-center-flag)))
	(setq cpt (+ 1 cpt))))))

;; A generic function which converts BHL source files.
(defun bhl2xxx (arg)
  "Convert the BHL source file into a HTML, SGML or LaTeX file.
ARG is the output format."
  (interactive)
  (bhl-convert-init arg)
  (unless (or (equal arg "txt") (not bhl-check-tags-flag))
    (bhl-check-existing-tags arg))
  (bhl-insert-preamble arg)
  (narrow-to-region (point) (point-max))
  (bhl-convert-misc-1 arg)
  (unless (or (equal arg "txt") (equal arg "latex"))
    (condition-case nil 
	(bhl-convert-para arg)
      (error (add-to-list 'bhl-conversion-log 
			  "BHL was unable to convert paragraphs properly."))))
  (bhl-convert-misc-2 arg)
  (bhl-update-toc)
  (bhl-skip-toc arg)
  (bhl-convert-sec-tag arg)
  (add-to-list 'bhl-conversion-log
	       (concat "Sections' numbering: "
		       (symbol-name bhl-numbered-section-flag) "."))
  (widen)
  (goto-char (point-max))
  (insert (cond ((equal arg "latex")  "\n\n\\end{document}")
		((equal arg "html")  "\n</BODY>\n</HTML>")
		((equal arg "sgml")  "</ARTICLE>")
		(t "")))
  (unless (equal arg "txt") (bhl-check-comment arg))
  (local-set-key "\C-c\C-o" 'bhl-view-log)
  (message "Converting into %s format...done. Type C-c C-o to view log." arg))

;;; bhl.el ends here

